!
! Contains: Three routines of EVP dynamics. The driving routine is EVPdynamics.
! 2D indices are used to distinguish between boundary and internal nodes.
! Standard EVP 
!============================================================================
subroutine stress_tensor
! EVP rheology implementation. Computes stress tensor components based on ice 
! velocity field. They are stored as elemental arrays (sigma11, sigma22 and
! sigma12). 

! Modification are now made (December 2014) which improve convergence, but 
! deviate from the standard EVP.  

use param
use iceparam
use mesh
use elements 
use ice
use parsup
implicit none

real(kind=8)   :: eps11, eps12, eps22, eta, xi, pressure, delta, eps1, eps2
integer        :: elem, elnodes(3),i
real(kind=8)   :: val3, asum, msum, vale, dx(3), dy(3)
real(kind=8)   :: det1, det2, r1, r2, r3, si1, si2, dte 
real(kind=8)   :: zeta, delta_inv, meancos, usum, vsum

  val3=1.0_8/3.0_8
  vale=1.0_8/(ellipse**2)
   
  dte=dt/(1.0_8*evp_rheol_steps)
  det1=1.0_8+0.5_8*Tevp_inv*dte
  det2=1.0_8+0.5_8*Tevp_inv*dte  !!!! Modified: (1) Should be with ellipse**2 
                                 !!!! factor in the standar implementation  
  det1=1.0_8/det1
  det2=1.0_8/det2
  
  do i=1,myDim_elem2D
     elem=myList_elem2D(i)     
     elnodes=elem2D_nodes(:,elem)
      ! ===== Check if there is ice on elem
     msum=sum(m_ice(elnodes))*val3
     if(msum<=0) cycle
     asum=sum(a_ice(elnodes))*val3
     
     dx=bafux(:,elem)
     dy=bafuy(:,elem)

     meancos=sin_elem2D(elem)/cos_elem2D(elem)/r_earth  !metrics
     vsum=sum(v_ice(elnodes))                           !metrics   
     usum=sum(u_ice(elnodes))                           !metrics

      ! ===== Deformation rate tensor on element elem:
     eps11=sum(dx*u_ice(elnodes))
     eps11=eps11-val3*vsum*meancos                !metrics
     eps22=sum(dy*v_ice(elnodes))
     eps12=0.5_8*sum(dy*u_ice(elnodes) + dx*v_ice(elnodes))
     eps12=eps12+0.5_8*val3*usum*meancos          !metrics        
      ! ======= Switch to eps1,eps2
     eps1=eps11+eps22
     eps2=eps11-eps22   
     
      ! ====== moduli:
     delta=eps1**2+vale*(eps2**2+4.0_8*eps12**2)
     delta=sqrt(delta) 
     pressure=pstar*(msum)*exp(-c_pressure*(1.0_8-asum))
      ! =======================================
      ! ===== Here the EVP rheology piece starts
      ! =======================================
     pressure=0.5_8*pressure
     delta_inv=1.0_8/(delta+delta_min) 
      zeta=pressure*delta_inv
      pressure=pressure*delta*delta_inv    ! Pressure substitution as in VP 
      pressure=pressure*Tevp_inv
      zeta=zeta*Tevp_inv
      				     
     r1=zeta*(eps11+eps22) - pressure
     r2=zeta*(eps11-eps22)*vale            !!!! Modified: (2) should be without vale 
     r3=zeta*eps12*vale                    !!!! in the standard case. (1) and (2)  
     si1=sigma11(elem)+sigma22(elem)
     si2=sigma11(elem)-sigma22(elem)
     
     si1=det1*(si1+dte*r1)
     si2=det2*(si2+dte*r2)
     sigma12(elem)=det2*(sigma12(elem)+dte*r3)
     sigma11(elem)=0.5_8*(si1+si2)
     sigma22(elem)=0.5_8*(si1-si2)
    end do
   
end subroutine stress_tensor

!===================================================================
subroutine stress2rhs
! EVP implementation:
! Computes the divergence of stress tensor and puts the result into the
! rhs vectors 

USE MESH
USE ELEMENTS
USE PARAM
USE ICEPARAM
USE ICE_THERM_PARMS
USE ice
USE PARSUP


IMPLICIT NONE
INTEGER      :: row, elem, elnodes(3), k, i  
REAL(kind=8) :: mass, aa
REAL(kind=8) :: cluster_area,elevation_elem(3)
REAL(kind=8) :: dx(3), dy(3), meancos, val3

val3=1/3.0_8
 DO i=1, myDim_nod2D
     row=myList_nod2D(i) 
     rhs_u(row)=0.0
     rhs_v(row)=0.0
     rhs_a(row)=0.0       ! these are used as temporal storage here
     rhs_m(row)=0.0       ! for the contribution due to ssh
 END DO
 
 do i=1,myDim_elem2D
     elem=myList_elem2D(i)    
     elnodes=elem2D_nodes(:,elem)
      ! ===== Skip if ice is absent
     aa=product(m_ice(elnodes))*product(a_ice(elnodes))
     if (aa==0) CYCLE
      ! =====

     meancos=sin_elem2D(elem)/cos_elem2D(elem)/r_earth         !metrics
     dx=bafux(:,elem)
     dy=bafuy(:,elem)
     elevation_elem=elevation(elnodes)
     
     DO k=1,3
        row=elnodes(k)
        rhs_u(row)=rhs_u(row) - voltriangle(elem) * &
             (sigma11(elem)*dx(k)+sigma12(elem)*(dy(k)) &
             +sigma12(elem)*val3*meancos)                          !metrics
        rhs_v(row)=rhs_v(row) - voltriangle(elem) * &
             (sigma12(elem)*dx(k)+sigma22(elem)*dy(k) &   
             -sigma11(elem)*val3*meancos)
     END DO
      ! use rhs_m and rhs_a for storing the contribution from elevation:
     aa=g*voltriangle(elem)/3.0_8
     DO k=1,3
        row=elnodes(k)
        rhs_a(row)=rhs_a(row)-aa*sum(dx*elevation_elem)	    
        rhs_m(row)=rhs_m(row)-aa*sum(dy*elevation_elem)
     END DO     
 end do 
 
  DO i=1, myDim_nod2D
     row=myList_nod2D(i)             
     cluster_area=lmass_matrix(row)
     mass=cluster_area*(m_ice(row)*rhoice+m_snow(row)*rhosno)
     
     if (mass>=1.0e-4) then
     rhs_u(row)=rhs_u(row)/mass + rhs_a(row)/cluster_area 
     rhs_v(row)=rhs_v(row)/mass + rhs_m(row)/cluster_area 
     else
     rhs_u(row)=0.  
     rhs_v(row)=0.
     end if
  END DO 
  
end subroutine stress2rhs
!===================================================================
subroutine EVPdynamics
! EVP implementation. Does cybcycling and boundary conditions.  
USE MESH
USE ELEMENTS
USE PARAM
USE ICE
USE ICEPARAM
USE ICE_THERM_PARMS
use PARSUP

IMPLICIT NONE
integer      :: steps, shortstep
real(kind=8) :: rdt
real(kind=8)    ::  drag, inv_mass, det, umod, rhsu, rhsv
integer         ::  i,j
real(kind=8)    :: ax, ay
    rdt=dt/(1.0_8*evp_rheol_steps)
    steps=evp_rheol_steps
    ax=cos(theta_io)
    ay=sin(theta_io)
do shortstep=1, steps 
 call stress_tensor
 call stress2rhs
 do j=1,myDim_nod2D 
    i=myList_nod2D(j)
  if (a_ice(i) > 0.01) then             ! If ice is present, update velocities
   inv_mass=(rhoice*m_ice(i)+rhosno*m_snow(i))/a_ice(i)
   inv_mass=max(inv_mass, 9.0)        ! Limit the weighted mass 
                                       ! if it is too small
   inv_mass=1.0_8/inv_mass

   umod=sqrt((u_ice(i)-u_w(i))**2+(v_ice(i)-v_w(i))**2)
   drag=C_d_io*umod*density_0*inv_mass
   
   rhsu=u_ice(i)+rdt*(drag*(ax*u_w(i)-ay*v_w(i))+inv_mass*stress_atmice_x(i)+rhs_u(i))
   rhsv=v_ice(i)+rdt*(drag*(ax*v_w(i)+ay*u_w(i))+inv_mass*stress_atmice_y(i)+rhs_v(i))

   det=(1.+ax*drag*rdt)**2+(rdt*coriolis_nod2D(i)+rdt*ay*drag)**2
   det=bc_index_nod2D(i)/det
   u_ice(i)=det*((1.0_8+ax*drag*rdt)*rhsu+rdt*(coriolis_nod2D(i)+ay*drag)*rhsv)
   v_ice(i)=det*((1.0_8+ax*drag*rdt)*rhsv-rdt*(coriolis_nod2D(i)+ay*drag)*rhsu)
   else                           ! Set ice velocity equal to water velocity
   u_ice(i)=0.0 !u_w(i)
   v_ice(i)=0.0 !v_w(i)
   end if
 end do
 call exchange_nod2D(u_ice)
 call exchange_nod2D(v_ice)    
 END DO
 


end subroutine EVPdynamics
!===================================================================

