!
! New evp implementation following Bouillion et al. 2013
!   


subroutine stress_tensor_b
  ! Internal stress tensor
  ! New implementation following Boullion et al, Ocean Modelling 2013.
  ! SD, 30.07.2014
  !===================================================================
  use param
  use iceparam
  use mesh
  use elements 
  use ice
  use parsup
  implicit none

  integer        :: i, elem, elnodes(3)
  real(kind=8)   :: dx(3), dy(3), msum, asum
  real(kind=8)   :: eps11, eps12, eps22, eps1, eps2, pressure, delta
  real(kind=8)   :: val3, meancos, usum, vsum, vale
  real(kind=8)   :: det1, det2, r1, r2, r3, si1, si2
  
  val3=1.0_8/3.0_8
  vale=1.0_8/(ellipse**2)
   
 
     det2=1.0_8/(1.0_8+alpha_evp)
     det1=alpha_evp*det2
     

  do i=1,myDim_elem2D
     elem=myList_elem2D(i)                     
     elnodes=elem2D_nodes(:,elem)

     msum=sum(m_ice(elnodes))*val3
     !if(msum<=0.01) cycle
     asum=sum(a_ice(elnodes))*val3
     
     dx=bafux(:,elem)
     dy=bafuy(:,elem)     
      ! METRICS:
          vsum=sum(v_ice_aux(elnodes))
          usum=sum(u_ice_aux(elnodes))
          meancos=sin_elem2D(elem)/cos_elem2D(elem)/r_earth
      !  

      ! ====== Deformation rate tensor on element elem:
     eps11=sum(dx*u_ice_aux(elnodes))
     eps11=eps11-val3*vsum*meancos                !metrics
     eps22=sum(dy*v_ice_aux(elnodes))
     eps12=0.5_8*sum(dy*u_ice_aux(elnodes) + dx*v_ice_aux(elnodes))
     eps12=eps12+0.5_8*val3*usum*meancos          !metrics 
     
      ! ======= Switch to eps1,eps2
     eps1=eps11+eps22
     eps2=eps11-eps22   
     
      ! ====== moduli:
     delta=eps1**2+vale*(eps2**2+4.0_8*eps12**2)
     delta=sqrt(delta)
    
     pressure=pstar*msum*exp(-c_pressure*(1.0_8-asum))/(delta+delta_min)
    
        r1=pressure*(eps1-delta) 
        r2=pressure*eps2*vale
        r3=pressure*eps12*vale
        si1=sigma11(elem)+sigma22(elem)
        si2=sigma11(elem)-sigma22(elem)

        si1=det1*si1+det2*r1
        si2=det1*si2+det2*r2
        sigma12(elem)=det1*sigma12(elem)+det2*r3
        sigma11(elem)=0.5_8*(si1+si2)
        sigma22(elem)=0.5_8*(si1-si2)
  end do
 ! Equations solved in terms of si1, si2, eps1, eps2 are (43)-(45) of 
 ! Boullion et al Ocean Modelling 2013, but in an implicit mode:
 ! si1_{p+1}=det1*si1_p+det2*r1, where det1=alpha/(1+alpha) and det2=1/(1+alpha),
 ! and similarly for si2 and sigma12
  
end subroutine stress_tensor_b

!
! ==================================================================
! 
subroutine ssh2rhs
  ! Compute the contribution from the elevation to the rhs
  ! S.D. 30.07.2014
USE MESH
USE ELEMENTS
USE PARAM
USE ICEPARAM
USE ICE_THERM_PARMS
USE ice
USE PARSUP
  implicit none
  
  integer       :: i ,k, row, elem, elnodes(3)
  real(kind=8)  :: dx(3), dy(3), vol
  real(kind=8)  :: val3, meancos, aa, bb
  real(kind=8)  :: elevation_elem(3)

  val3=1.0_8/3.0_8
  
  ! use rhs_m and rhs_a for storing the contribution from elevation:
  do i=1, myDim_nod2d 
     row=myList_nod2D(i)
     rhs_a(row)=0.0
     rhs_m(row)=0.0
  end do

  do i=1,myDim_elem2d         
     elem=myList_elem2d(i)
     elnodes=elem2D_nodes(:,elem)
     vol=voltriangle(elem)
     dx=bafux(:,elem)
     dy=bafuy(:,elem)     
     elevation_elem=elevation(elnodes)
     bb=g*val3*vol
     aa=bb*sum(dx*elevation_elem)
     bb=bb*sum(dy*elevation_elem)
        rhs_a(elnodes)=rhs_a(elnodes)-aa	    
        rhs_m(elnodes)=rhs_m(elnodes)-bb
  end do
  
  
  
end subroutine ssh2rhs
!
!===================================================================
!
subroutine stress2rhs_b

  ! add internal stress and ssh gradient terms to the rhs
  ! SD, 30.07.2014
  !-----------------------------------------------------------------
  
USE MESH
USE ELEMENTS
USE PARAM
USE ICEPARAM
USE ICE_THERM_PARMS
USE ice
USE PARSUP
  implicit none
  
  integer       :: i ,k, row, elem, elnodes(3)
  real(kind=8)  :: dx(3), dy(3), vol
  real(kind=8)  :: val3, meancos, aa, bb
  real(kind=8)  :: mass, cluster_area, elevation_elem(3)

  val3=1.0_8/3.0_8
  
  do i=1, myDim_nod2d 
     row=myList_nod2D(i)
     rhs_u(row)=0.0
     rhs_v(row)=0.0
  end do

  do i=1,myDim_elem2d         
     elem=myList_elem2d(i)
     elnodes=elem2D_nodes(:,elem)
    ! if(sum(a_ice(elnodes))<0.01) cycle
     
     vol=voltriangle(elem)
     dx=bafux(:,elem)
     dy=bafuy(:,elem)     
     meancos=sin_elem2D(elem)/cos_elem2D(elem)/r_earth    !metrics

     do k=1,3
        row=elnodes(k)
        rhs_u(row)=rhs_u(row) - vol* &
             (sigma11(elem)*dx(k)+sigma12(elem)*dy(k))    &
       -vol*sigma12(elem)*val3*meancos                    !metrics
        rhs_v(row)=rhs_v(row) - vol* &
             (sigma12(elem)*dx(k)+sigma22(elem)*dy(k))    &
       +vol*sigma11(elem)*val3*meancos                    ! metrics
     end do
  end do
  
  do i=1, myDim_nod2d               
     row=myList_nod2D(i)
     cluster_area=lmass_matrix(row)
     mass=(m_ice(row)*rhoice+m_snow(row)*rhosno)
     mass=mass/(1.0_8+mass*mass)
        rhs_u(row)=(rhs_u(row)*mass + rhs_a(row))/cluster_area 
        rhs_v(row)=(rhs_v(row)*mass + rhs_m(row))/cluster_area 
  end do

end subroutine stress2rhs_b
!
!===================================================================
!
subroutine stress2rhs_b_old

  ! add internal stress and ssh gradient terms to the rhs
  ! Slightly slower because the gradient of elevation is computed on 
  ! each subcycling step.  
  !-----------------------------------------------------------------
  
USE MESH
USE ELEMENTS
USE PARAM
USE ICEPARAM
USE ICE_THERM_PARMS
USE ice
USE PARSUP
  implicit none
  
  integer       :: i ,k, row, elem, elnodes(3)
  real(kind=8)  :: dx(3), dy(3), vol
  real(kind=8)  :: val3, meancos, aa, bb
  real(kind=8)  :: mass, cluster_area, elevation_elem(3)

  val3=1.0_8/3.0_8
  
  do i=1, myDim_nod2d 
     row=myList_nod2D(i)
     rhs_u(row)=0.0
     rhs_v(row)=0.0
     rhs_a(row)=0.0
     rhs_m(row)=0.0
  end do

  do i=1,myDim_elem2d         
     elem=myList_elem2d(i)
     elnodes=elem2D_nodes(:,elem)

     if(product(m_ice(elnodes))*product(a_ice(elnodes))==0) cycle
     
     vol=voltriangle(elem)
     dx=bafux(:,elem)
     dy=bafuy(:,elem)     
     elevation_elem=elevation(elnodes)
      meancos=sin_elem2D(elem)/cos_elem2D(elem)/r_earth   !metrics

     do k=1,3
        row=elnodes(k)
        rhs_u(row)=rhs_u(row) - vol* &
             (sigma11(elem)*dx(k)+sigma12(elem)*dy(k))    &
       -vol*sigma12(elem)*val3*meancos                    !metrics
        rhs_v(row)=rhs_v(row) - vol* &
             (sigma12(elem)*dx(k)+sigma22(elem)*dy(k))    &
       +vol*sigma11(elem)*val3*meancos                    ! metrics
     end do

     ! use rhs_m and rhs_a for storing the contribution from elevation:
     bb=g*val3*vol
     aa=bb*sum(dx*elevation_elem)
     bb=bb*sum(dy*elevation_elem)
        rhs_a(elnodes)=rhs_a(elnodes)-aa	    
        rhs_m(elnodes)=rhs_m(elnodes)-bb
  end do
  
  do i=1, myDim_nod2d               
     row=myList_nod2D(i)
     cluster_area=lmass_matrix(row)
     mass=cluster_area*(m_ice(row)*rhoice+m_snow(row)*rhosno)

     if (mass>1.0e-4) then
        rhs_u(row)=rhs_u(row)/mass + rhs_a(row)/cluster_area 
        rhs_v(row)=rhs_v(row)/mass + rhs_m(row)/cluster_area 
     else
        rhs_u(row)=0.  
        rhs_v(row)=0.
     end if
  end do

  ! Clean neighbours: Should not be needed
  do i=myDim_nod2d+1,eDim_nod2d+myDim_nod2d  
     row=myList_nod2D(i)
     rhs_u(row)=0. 
     rhs_v(row)=0.
  end do

end subroutine stress2rhs_b_old
!
!===================================================================
!
subroutine EVPdynamics_b
  ! assemble rhs and solve for ice velocity
  ! New implementation based on Bouillion et al. Ocean Modelling 2013
  ! SD 30.07.14
  !---------------------------------------------------------

USE MESH
USE ELEMENTS
USE PARAM
USE ICE
USE ICEPARAM
USE ICE_THERM_PARMS
use PARSUP

  implicit none
  integer         :: steps, shortstep, i, j
  real(kind=8)    :: rdt, drag, det, fc
  real(kind=8)    :: thickness, inv_thickness, umod, rhsu, rhsv
  REAL(kind=8)    :: t0,t1, t2, t3, t4, t5, t00, txx
 
  steps=evp_rheol_steps
  
  u_ice_aux=u_ice    ! Initialize solver variables
  v_ice_aux=v_ice
  call ssh2rhs
  do shortstep=1, steps 
     call stress_tensor_b
     call stress2rhs_b
     do j=1,myDim_nod2D 
     i=myList_nod2D(j) 
         thickness=(rhoice*m_ice(i)+rhosno*m_snow(i))/max(a_ice(i),0.01)
         thickness=max(thickness, 9.0)   ! Limit if it is too small (0.01 m)
         inv_thickness=1.0_8/thickness

         umod=sqrt((u_ice_aux(i)-u_w(i))**2+(v_ice_aux(i)-v_w(i))**2)
         drag=dt*C_d_io*umod*density_0*inv_thickness

         !rhs for water stress, air stress, and rhs_u/v (internal stress + ssh)
         rhsu=u_ice(i)+drag*u_w(i)+dt*(inv_thickness*stress_atmice_x(i)+rhs_u(i))
         rhsv=v_ice(i)+drag*v_w(i)+dt*(inv_thickness*stress_atmice_y(i)+rhs_v(i))

         rhsu=beta_evp*u_ice_aux(i)+rhsu
	 rhsv=beta_evp*v_ice_aux(i)+rhsv
         !solve (Coriolis and water stress are treated implicitly)
         fc=dt*coriolis_nod2D(i)
         det=(1.0_8+beta_evp+drag)**2+fc**2
         det=bc_index_nod2D(i)/det
         u_ice_aux(i)=det*((1.0+beta_evp+drag)*rhsu+fc*rhsv)
         v_ice_aux(i)=det*((1.0+beta_evp+drag)*rhsv-fc*rhsu)
     end do
     
     call exchange_nod2D(u_ice_aux)
     call exchange_nod2D(v_ice_aux)
  end do
  
    do j=1, myDim_nod2D+eDim_nod2D   ! Added 28.10.14 for full compatibility with 
    i=myList_nod2D(j)                ! the VP solver 
    if(a_ice(i)<=0.01) then
    u_ice_aux(i)=0.0
    v_ice_aux(i)=0.0
    end if
    end do 
    
    u_ice=u_ice_aux
    v_ice=v_ice_aux
 
end subroutine EVPdynamics_b
