! =====================
!  Sea ice model
!  Version 2x. New features added:
!  VP solver, FCT advection and modified EVP solver by Bouillon (2013).  
!  Questions to S. Danilov (dynamics) and Q. Wang and R. Timmermann 
!  (thermodynamics). 
!  Subject to license agreement.
!  sergey.danilov@awi.de, 2014
! ======================

! This file contails all modules, subroutine feim_step (ice model
! step) and the main program. The modules PARAM, data_TYPES, MESH,
! ELEMENTS and PARSUP are derived from
! FESOM, the coupled version relies on full FESOM modules.
! ======================
!
! ===============================================================
!
module PARAM
 ! Defines some basic parameters
  implicit none
  save
  real(kind=8), parameter  :: pi=3.141592653589793, rad=pi/180.0
  real(kind=8), parameter  :: omega=2.0*pi/(24.0*60.0*60.0)
  real(kind=8)             :: g=9.806                               ![m/s^2]
  real(kind=8), parameter  :: r_earth =6.3675e6                     ![m]
  real(kind=8), parameter  :: density_0=1027.7                      ![kg/m^3]
  real(kind=8)             :: dt=3600./2.              ! time step
  integer(KIND=4)          :: iter
  !
  logical                  :: lfirst=.true.            ! used in solver to set structure  
  logical                  :: r_restart=.false.        !
  logical                  :: cartesian=.false.        ! mesh is spherical if .false.
  !
  logical                  :: cyclic_geometry=.true.   !.true. for periodical geometry
  real(kind=8)             :: domain_length=20.0*rad   ! unit in rad.
  !
  !the mesh is stored there:  
  !
  character*100            :: meshpath='./mesh/'
  
end module PARAM
!
!===============================================================
!
module data_TYPES
  !  
  ! Defines the structures to keep sparse matrices (VP solver)
  ! + neighborhood information
  ! 
  implicit none
  save
  type sparse_matrix 
     integer :: nza
     integer :: dim
     real(kind=8), pointer, dimension(:)      :: values
     integer(KIND=4), pointer,   dimension(:) :: colind
     integer(KIND=4), pointer,   dimension(:) :: rowptr
  end type sparse_matrix
  type addresstype
     integer                                  :: nmb
     integer(KIND=4), dimension(:), pointer   :: addresses
  end type addresstype
end module data_TYPES
!
! =============================================================
!
module MESH
  !
  ! Arrays used to keep the mesh information
  ! 
  use data_TYPES
  implicit none
  save
  !
  integer                                      :: nod2D          ! the number of nodes        
  real(kind=8), allocatable, dimension(:,:)    :: coord_nod2D    ! node coordinates
  real, allocatable, dimension(:)              :: bc_index_nod2D ! 0 at boundaries 
  real(kind=8), allocatable, dimension(:)      :: cos_elem2D     
  real(kind=8), allocatable, dimension(:)      :: sin_elem2D 
  integer, allocatable, dimension(:)           :: col_pos        ! auxiliary array
  !  
  type(addresstype), allocatable, dimension(:) :: nod_in_elem2D  ! elems containing a node   
  type(addresstype), allocatable, dimension(:) :: nghbr_nod2D    ! nodes neighbor. a node
  !
end module MESH
!
! =============================================================
!
module ELEMENTS
  !
  ! Arrays defined here are also used to keep mesh information
  ! Separation between MESH and ELEMENTS is for 
  ! historical reasons (and is inherited from FESOM)
  ! 
  implicit none
  save
  integer                                      :: elem2D        ! the number of elements 
  integer(KIND=4), allocatable, dimension(:,:) :: elem2D_nodes  ! their indices
  real(kind=8)                                 :: Vol2D         ! aux. quantity
  real(kind=8), allocatable, dimension(:,:)    :: derivative_stdbf  ! aux. arrays
  real(kind=8), allocatable, dimension(:,:)    :: bafux, bafuy  ! derivatives of test
                                                                ! functions 
  real(kind=8), allocatable, dimension(:)      :: voltriangle   ! areas of elements    
end module ELEMENTS
!
! =============================================================
!
module PARSUP
  !
  ! Arrays used to organize the parallel support
  !
  implicit none
  save 

#ifdef PETSC
#include "finclude/petsc.h"
#else
  include 'mpif.h'
#endif
  integer      :: maxPEnum              ! max number of cores
  type com_struct                       ! structures to store send and receive node data
     integer    :: rPEnum
     integer, dimension(:), pointer :: rPE
     integer, dimension(:), pointer :: rptr
     integer, dimension(:), pointer :: rlist
     integer    :: sPEnum
     integer, dimension(:), pointer :: sPE
     integer, dimension(:), pointer :: sptr
     integer, dimension(:), pointer :: slist
  end type com_struct

  type(com_struct)   :: com_nod2D
  ! Buffer arrays to store information to be communicated
  type com_array
     real(kind=8), dimension(:), pointer :: array
  end  type com_array
  ! 2D buffer of FESOM is used
  type(com_array), allocatable             :: s_buff_ssh(:), r_buff_ssh(:) 
                                         
  ! general MPI part
  integer            :: MPIERR
  integer            :: npes
  integer            :: mype
  integer, allocatable, dimension(:)  :: part       

  ! Mesh partition
  integer                             :: myDim_nod2D, eDim_nod2D 
  integer, allocatable, dimension(:)  :: myList_nod2D
  integer                             :: myDim_elem2D
  integer, allocatable, dimension(:)  :: myList_elem2D
  end module PARSUP
!
!===========================================================================
!
MODULE ICEPARAM
  !
  ! Ice specific parameters
  !
  IMPLICIT NONE
  SAVE

  REAL(kind=8), ALLOCATABLE, DIMENSION(:)    :: coriolis_nod2D  
 ! ice model parameters:
  ! RHEOLOGY
  REAL(kind=8), PARAMETER  :: Pstar = 27500             !15000 !was 20000.   ![N/m^2]
  REAL(kind=8), PARAMETER  :: ellipse =2.               !
  REAL(kind=8), PARAMETER  :: c_pressure =20.0          !
  REAL(kind=8)             :: delta_min=2.0e-9          ! [s^(-1)]
  INTEGER                  :: evp_rheol_steps=120       ! EVP rheology
                                                        ! cybcycling steps
  REAL(kind=8)             :: ice_gamma_fct=0.25        ! smoothing parameter
                                                        ! in ice fct advection
  LOGICAL                  :: ice_VP_rheology=.true.    ! VP=.true., EVP=.false.
  LOGICAL                  :: ice_EVP1=.false.          ! use standard EVP if .true.
  integer                  :: ice_VP_iter=1             ! Pickard iterations >=1
  REAL(kind=8)             :: ice_VP_soltol=1.0e-8      ! tolerance in PETCs  
  REAL(kind=8)             :: Tevp_inv                  ! standard EVP parameter  
  integer                  :: alpha_evp=100, beta_evp=100       
  integer                  :: num_iter_solve=3          ! Iterations of mass matrix solver
  real(kind=8)             :: theta_io                  ! rotation angle
                                                        ! (ice-ocean)
  REAL(kind=8), PARAMETER  :: C_d_io= 0.0055   !sd  was 0.003    ! ice-water drag coefficient
  REAL(kind=8), PARAMETER  :: rho_ice=900
  REAL(kind=8), PARAMETER  :: rho_snow=300
END MODULE ICEPARAM

!=============================================================================

MODULE ICE
!
! Arrays used to store ice variables and organize coupling
!

USE data_types
implicit none
save
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: u_ice, v_ice, m_ice, a_ice 
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: u_ice_aux, v_ice_aux 
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: rhs_u, rhs_v, m_snow
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: rhs_m, rhs_a, rhs_ms
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: u_w, v_w
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: elevation
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: mass_matrix
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: lmass_matrix
  TYPE(sparse_matrix)                             :: icestiff
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: sigma11, sigma12, sigma22  
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: fresh_wa_flux
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: net_heat_flux
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: S_oc_array, T_oc_array
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: stress_iceoce_x         
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: stress_iceoce_y
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: stress_atmice_x         
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: stress_atmice_y
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: stress_atmoce_x         
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: stress_atmoce_y
  ! auxiliary arrays required for fct
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: m_icel, a_icel, m_snowl
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: dm_ice, da_ice, dm_snow
  REAL(kind=8), ALLOCATABLE, DIMENSION(:,:)       :: icefluxes
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: icepplus, icepminus
END MODULE ICE

!=====================================================================

MODULE ATM_FORCING
USE data_types
implicit none
save
    
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: u_wind, v_wind 
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: shortwave, longwave
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: Tair, Tdew
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: Pair, precipitation
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: evaporation
  REAL(kind=8), ALLOCATABLE, DIMENSION(:)         :: shum, cloudiness  !RTnew


END MODULE ATM_FORCING

!=====================================================================

MODULE ice_therm_parms
implicit none
real*8, parameter :: rhoair=  1.3    ! Air density
real*8, parameter :: rhowat= 1025.   ! Water density
real*8, parameter :: rhoice=  910.   ! Ice density
real*8, parameter :: rhosno=  290.   ! Snow density

real*8, parameter :: sice = 5.0      ! Ice salinity 3.2--5.0 ppt.

integer, parameter :: iclasses=7     ! Number of ice thickness gradations for ice growth calcs.
real*8, parameter :: h0=1.0	     ! Lead, closing parameter 0.5 [m] standard
real*8, parameter :: Saterm=0.5      ! Sa - term parameter 0.5 [m] standard
real*8, parameter :: hmin= 0.05      ! Cut-off ice thickness
real*8, parameter :: Armin=0.15      ! Minimum ice concentration

real*8, parameter :: tmelt=273.16    ! 0 deg C expressed in K 
real*8, parameter :: cc=4.20E6       ! Volumetr. heat cap. of water [J/m**3/K](cc = rhoice*cp)
real*8, parameter :: cl=3.02E8       ! Volumetric latent heat of sea ice [J/m**3] 
real*8, parameter :: q0=1./cl        ! 1/volumetric heat of fusion
real*8, parameter :: cpair=1004.     ! Specific heat of air [J/(kg * K)] 
real*8, parameter :: clhw=2.500E6    ! Specific latent heat [J/kg]: water	-> water vapor;
real*8, parameter :: clhi=2.834E6    !                              sea ice	-> water vapor
real*8, parameter :: cdsens=1.75E-3  ! Bulk sensible heat transfer coefficient, SIOM standard 
real*8, parameter :: cdlat =1.75E-3  ! Bulk latent heat transfer coefficients,  SIOM standard 

!RT:
real*8, parameter :: gamma_t=1.e-4   ! heat transfer coefficient ocean -> ice
!RT-

real*8, parameter :: qsw=17.2694     ! Constants for latent heat fluxes  
real*8, parameter :: tqw=237.3       ! over water
real*8, parameter :: qsi=21.8746     ! and ice
real*8, parameter :: tqi=265.5

real*8, parameter :: d1=rhoair*cpair*cdsens  ! coefficients for bulk formulas for heat fluxes
real*8, parameter :: d2w=rhoair*clhw*cdlat
real*8, parameter :: d2i=rhoair*clhi*cdlat

real*8, parameter :: Cd_thrm_i_o=1.0e-3  ! Ocean-Ice thermoconductivity coefficient

real*8, parameter :: emiss=0.97          ! Emissivity of Snow/Ice

real*8, parameter :: boltzmann=5.67E-8   ! S. Boltzmann const.*longw. emissivity
real*8, parameter :: d3=boltzmann*emiss  ! SIOM standard (MH)  
real*8, parameter :: con   = 2.1656	 ! Thermal conductivities: ice;
real*8, parameter :: consn = 0.31	 !                         snow

!
real*8, parameter :: albsn=	0.85     ! Albedo: frozen snow
real*8, parameter :: albsnm=    0.75     !         melting snow
real*8, parameter :: albi=	0.75     !         frozen ice
real*8, parameter :: albm=	0.66     !         melting ice
real*8, parameter :: albw=	0.10     !         open water
!

END MODULE ice_therm_parms

!=======================================================================

MODULE ice_dyn_parms
!introduced by Ralph Timmermann, 17.9.2004

implicit none
real*8, parameter :: cdwin = 2.25e-3   ! drag coefficient atmosphere 
real*8, parameter :: cdwat = 5.00e-3   ! drag coefficient ocean - ice
real*8, parameter :: cdao  = 1.20e-3   ! drag coefficient atmosphere - ocean 

END MODULE ice_dyn_parms
!==============================================================================
subroutine feim_step
! 
! Sea ice model time step
!
use param
use mesh
use elements 
use ice
use iceparam
use parsup
use atm_forcing
implicit none

integer  :: time, globaltime, n, i
integer  :: year, month, day, globalday
real(kind=8) :: tmass
REAL(kind=8) :: t0,t1, t2, t3
t0=MPI_Wtime()

 ! ============
 ! Dynamics
 ! ============
 if(ice_VP_rheology) then
    call VPdynamics            ! VP solver is called
 elseif(ice_evp1) then
    call EVPdynamics           ! standard EVP
 else
    call EVPdynamics_b	       ! modified EVP	 	    
 end if
 
 t2=MPI_Wtime() 
 ! ============
 ! Advection 
 ! ============   
         call ice_TG_rhs
         call fct_ice_solve
         call cut_off
 ! ============ 
 ! Ice thermodynamic part
 ! ============
        !call thermodynamics   ! If uncommented, input has to be provided
	                       ! for the fields declared in modules ICE and ATM_FORCING 
 t1=MPI_Wtime()
   if (mype==0) then
      write(*,*) 'FEIM step took ', t1-t0
      write(*,*) 'FEIM dynamics  ', t2-t0
   endif

end subroutine feim_step
!==============================================================================
! Sea ice main routine 

program ice_main
USE iceparam
USE param
USE PARSUP
use ice
implicit none
integer      :: steps, step
  ! set the parameters
  !
  ! ######## THIS is contained in namelists of the coupled version
  !
  meshpath='/iblade/home/sdanilov/ice/mesh2/'
  dt=3600.0/2.0        ! time step
  
  evp_rheol_steps=120  ! the number of sybcycling steps in EVP
  delta_min=2.0e-9     ! (1/s) Limit for minimum divergence (Hibler, Hunke
                       ! normally use 2.0e-9, which does much stronger
		       ! limiting; valid for both VP and EVP 
  theta_io=0. !0.436   ! ice/ocean rotation angle. Implemented in EVP, VP, not in EVP_new
  ice_gamma_fct=0.5    ! smoothing parameter in ice fct advection
  ice_VP_rheology=.false.  ! VP=.true., EVP=.false.
  ice_VP_iter=10       ! The numberof Pickard iterations
  ice_EVP1=.true.      ! standard EVP .true., modified EVP .false.
  ice_VP_soltol=1.0e-10
   
  steps= 24*10*6       ! Total number of steps
  alpha_evp=500        ! Parameters  of modified EVP formulation in Bouillon (2013)
  beta_evp=1000
  Tevp_inv=4.0/dt      ! Time scale of the standard EVP
  evp_rheol_steps=1000 ! The number of subcycling time steps 
  
  !		       	    
  ! ########     
  !			    
    
  call par_init
          
  ! ######## 
  ! Standing-alone model requires setting up the mesh
  ! In the coupled version this set of routines is called 
  ! in the ocean model
  
  call set_mesh 
  !########
  call icestiff_matrix         ! We need it here to provide the connectivity
                               ! pattern for partit (called in set_par_support). 
			       ! In the coupled version, 
			       ! this call is in array_setup_ice
			             
  !########			       
  call set_par_support			   
  
  ! This is the ice specific part. 
  ! Instead of real io routines
  ! dynamical fields and forcing are initialized and set by 
  ! init_fields, read_restart_ice and update_forcing routines. 
  ! A user has to provide
  ! appropriate versions of those. 
  ! In a coupled version
  ! FESOM routines are used  
  ! ########
  call array_setup_ice
  ! ########
  call init_fields
  if(r_restart) then
     call read_restart_ice
     write(*,*) 'restart'
  end if    
  write(*,*) 'fields are initialized'
  ! ==========
  ! Time stepping
  ! ==========
  do step=1, steps
     if(mype==0)  write(*,*) 'DAY ', step*dt/(24*3600) 
      call update_forcing(step)
      call feim_step
      lfirst=.false.
  end do
  call ice_out
  call write_restart_ice
  call par_ex
end program ice_main
